#ifndef ENGINES_FRET_MODEL_WRAPPER_H
#define ENGINES_FRET_MODEL_WRAPPER_H

#include "../../types.h"
#include "../../storage/segmented_vector.h"
#include "../../cli/macros.h"

#include <memory>
#include <vector>

namespace fd {
namespace storage {
class CompressedArrayHashSet;
}
namespace sas_task {
class ProbabilisticSASTask;
}

namespace engines {
namespace fret {

template<typename Model>
class FretModel {
public:
    FretModel(std::shared_ptr<Model> model);
    ~FretModel() = default;

    State get_initial_state();

    void get_applicable_actions(std::vector<Action>& aops, const State& state);

    void get_successors(
            std::vector<std::pair<ValueType, State> >& successors,
            const State& state,
            const Action& action);

    void print_statistics() const;

    const std::shared_ptr<sas_task::ProbabilisticSASTask> get_underlying_task() const;
    const storage::CompressedArrayHashSet& get_state_cache() const;

    std::shared_ptr<Model> get_underlying_model() const;

    void fill_probabilities(
            std::vector<std::pair<ValueType, State> >& successors,
            const State& state,
            const Action& action);
    unsigned get_num_outcomes(const State& state, const Action& action);

    template<typename StateIterator>
    State add_trap(StateIterator begin, StateIterator end);
private:
    using TrapId = unsigned int;
    static const TrapId SELF;
    struct Trap {
        Trap(const State& repr)
            : repr(repr)
        {}
        Trap(const Trap& trap) = default;
        Trap&
        operator=(Trap&& other)
        {
            repr = other.repr;
            states = std::move(other.states); other.states.clear();
            applicable = std::move(other.applicable); other.applicable.clear();
            return *this;
        }
        void
        add_applicable(const State& state, const std::vector<Action>& aops)
        {
            applicable.reserve(applicable.size() + aops.size());
            for (unsigned i = 0; i < aops.size(); i++) {
                applicable.emplace_back(state, aops[i]);
            }
        }
        void
        merge_from(Trap& t)
        {
            assert(this != &t);
            states.insert(states.end(), t.states.begin(), t.states.end());
            t.states.clear();

            applicable.reserve(applicable.size() + t.applicable.size());
            for (unsigned i = 0; i < t.applicable.size(); i++) {
                applicable.emplace_back(std::move(t.applicable[i]));
            }
            t.applicable.clear();
        }
        void
        release_memory()
        {
            states.clear(); states.shrink_to_fit();
            applicable.clear(); applicable.shrink_to_fit();
        }

        State repr;
        std::vector<State> states;
        std::vector<std::pair<State, Action> > applicable;
    };
    TrapId& get_trap_id(const State& state);
    State translate_state(const State& state);

    std::shared_ptr<Model> m_underlying_model;
    storage::SegmentedVector<TrapId> m_trap_ids;
    storage::SegmentedVector<Trap> m_traps;

    unsigned m_stat_unused_traps;
};


template<typename Model>
FretModel<Model>::FretModel(
        std::shared_ptr<Model> model)
    : m_underlying_model(model)
{
    m_stat_unused_traps = 0;
}

template<typename Model>
State
FretModel<Model>::get_initial_state()
{
    return translate_state(m_underlying_model->get_initial_state());
}

template<typename Model>
void
FretModel<Model>::get_applicable_actions(
        std::vector<Action>& aops,
        const State& state)
{
    TrapId tid = get_trap_id(state);
    if (tid == SELF) {
        m_underlying_model->get_applicable_actions(aops, state);
    } else {
        const Trap& comp = m_traps[tid];
        aops.reserve(aops.size() + comp.applicable.size());
        for (Action i = 0; i < comp.applicable.size(); i++) {
            aops.push_back(i);
        }
    }
}

template<typename Model>
void
FretModel<Model>::get_successors(
        std::vector<std::pair<ValueType, State> >& successors,
        const State& state,
        const Action& action)
{
    TrapId tid = get_trap_id(state);
    int end = successors.size();
    if (tid == SELF) {
        m_underlying_model->get_successors(successors, state, action);
    } else {
        const Trap& comp = m_traps[tid];
        assert(action < comp.applicable.size());
        const std::pair<State, Action>& real_state_action = comp.applicable[action];
        m_underlying_model->get_successors(successors, real_state_action.first, real_state_action.second);
    }
    for (int i = successors.size() - 1; i >= end; i--) {
        successors[i].second = translate_state(successors[i].second);
    }
}

template<typename Model>
void
FretModel<Model>::print_statistics() const
{
    m_underlying_model->print_statistics();
    FD_LOG_MESSAGE("Traps: " << m_traps.size());
    FD_LOG_MESSAGE("Unused traps: " << m_stat_unused_traps);
}

template<typename Model>
const std::shared_ptr<sas_task::ProbabilisticSASTask>
FretModel<Model>::get_underlying_task() const
{
    return m_underlying_model->get_underlying_task();
}

template<typename Model>
const storage::CompressedArrayHashSet&
FretModel<Model>::get_state_cache() const
{
    return m_underlying_model->get_state_cache();
}

template<typename Model>
std::shared_ptr<Model>
FretModel<Model>::get_underlying_model() const
{
    return m_underlying_model;
}

template<typename Model>
void
FretModel<Model>::fill_probabilities(
        std::vector<std::pair<ValueType, State> >& successors,
        const State& state,
        const Action& action)
{
    TrapId tid = get_trap_id(state);
    if (tid == SELF) {
        m_underlying_model->fill_probabilities(successors, state, action);
    } else {
        const Trap& comp = m_traps[tid];
        assert(action < comp.applicable.size());
        const std::pair<State, Action>& real_state_action = comp.applicable[action];
        m_underlying_model->fill_probabilities(successors, real_state_action.first, real_state_action.second);
    }
}

template<typename Model>
unsigned
FretModel<Model>::get_num_outcomes(
        const State& state,
        const Action& action)
{
    TrapId tid = get_trap_id(state);
    if (tid == SELF) {
        return m_underlying_model->get_num_outcomes(state, action);
    } else {
        const Trap& comp = m_traps[tid];
        assert(action < comp.applicable.size());
        const std::pair<State, Action>& real_state_action = comp.applicable[action];
        return m_underlying_model->get_num_outcomes(real_state_action.first, real_state_action.second);
    }
}

template<typename Model>
template<typename StateIterator>
State
FretModel<Model>::add_trap(StateIterator begin, StateIterator end)
{
    assert(begin != end);

    static std::vector<Action> aops;
    assert(aops.empty());

    TrapId tid = m_traps.size();
    m_traps.push_back(Trap(*begin));
    Trap& trap = m_traps[tid];

    for (auto it = begin; it != end; it++) {
        TrapId& state_tid = get_trap_id(*it);
        if (state_tid == tid) {
            continue;
        }
        if (state_tid == SELF) {
            trap.states.push_back(*it);
            m_underlying_model->get_applicable_actions(aops, *it);
            trap.add_applicable(*it, aops);
            aops.clear();
            state_tid = tid;
            if (*it < trap.repr) {
                trap.repr = *it;
            }
            assert(get_trap_id(*it) == tid);
        } else {
            Trap& state_current_trap = m_traps[state_tid];
            for (int i = state_current_trap.states.size() - 1; i >= 0; i--) {
                const State& s = state_current_trap.states[i];
                m_trap_ids[s] = tid;
                if (s < trap.repr) {
                    trap.repr = s;
                }
            }
            trap.merge_from(state_current_trap);
            state_current_trap.release_memory();
        }
    }

    static std::vector<std::pair<ValueType, State> > successors;
    assert(successors.empty());

    unsigned i = 0;
    for (unsigned j = 0; j < trap.applicable.size(); j++) {
        m_underlying_model->get_successors(
                successors,
                trap.applicable[j].first,
                trap.applicable[j].second);
        bool self_loop = true;
        for (int k = successors.size() - 1; k >= 0; k--) {
            if (get_trap_id(successors[k].second) != tid) {
                self_loop = false;
                break;
            }
        }
        successors.clear();
        if (!self_loop) {
            if (i != j) {
                trap.applicable[i].swap(trap.applicable[j]);
            }
            i++;
        }
    }
    trap.applicable.resize(i);
    trap.applicable.shrink_to_fit();

    return trap.repr;
}

template<typename Model>
typename FretModel<Model>::TrapId&
FretModel<Model>::get_trap_id(const State& state)
{
    if (state >= m_trap_ids.size()) {
        m_trap_ids.resize(state + 1, SELF);
    }
    return m_trap_ids[state];
}

template<typename Model>
State
FretModel<Model>::translate_state(const State& state)
{
    TrapId tid = get_trap_id(state);
    if (tid == SELF) {
        return state;
    } else {
        return m_traps[tid].repr;
    }
}

template<typename Model>
const unsigned FretModel<Model>::SELF = -1;

}
}
}

#endif
