#ifndef ENGINE_HEURISTIC_DEPTH_FIRST_SEARCH_H
#define ENGINE_HEURISTIC_DEPTH_FIRST_SEARCH_H

#include "engine.h"
#include "../types.h"
#include "../storage/state_map.h"
#include "../storage/segmented_vector.h"
#include "../heuristics/heuristic.h"
#include "../heuristics/evaluation_result.h"
#include "../utils/rng.h"

#include <memory>
#include <deque>
#include <cassert>
#include <algorithm>

namespace fd {
namespace engines {
namespace hdfs {

template<typename Model, typename Reward>
class HeuristicDepthFirstSearch : public Engine<Model, Reward>
{
public:
    HeuristicDepthFirstSearch(
        std::shared_ptr<Model> t_model,
        std::shared_ptr<Reward> r_model,
        std::shared_ptr<heuristics::Heuristic<Reward> > heuristic = nullptr,
        bool forward_updates = true,
        bool backward_updates = true,
        bool stop_at_inconsistent = true,
        bool complete_policy_exploration = true);

    void reset_and_prepare();
    void prepare();
    ValueType get_reward(const State& state);
    Action get_policy_action(const State& state);

    bool update_reward_get_policy(Action& p, const State& state);
    bool update_reward(const State& state);

private:
    struct PerStateInfo {
        PerStateInfo();
        
        bool is_new() const;
        bool is_solved() const;
        bool is_terminal() const;

        void set_opened();
        void set_solved();
        void set_terminal();

        void unset_solved();

        ValueType reward;
        Action policy;
    private:
        unsigned status;

        static const unsigned OPENED;
        static const unsigned SOLVED;
        static const unsigned TERMINAL;
    };
    
    struct SccInfo {
        static const unsigned UNDEFINED;
        SccInfo() : index(UNDEFINED), lowlink(UNDEFINED), onstack(false) {}
        unsigned index;
        unsigned lowlink;
        bool onstack;
    };

    struct ExpansionInfo {
        ExpansionInfo(const State& state)
            : state(state), flag(false) {}
        State state;
        bool flag;
        std::vector<std::pair<ValueType, State> > successors;
    };

    PerStateInfo& lookup_info(const State& state);
    std::pair<bool, bool> update_reward(
            std::vector<std::pair<ValueType, State> >& policy_successors,
            PerStateInfo& info,
            const State& state);
    void sort_successors_for_exploration(
            std::vector<std::pair<ValueType, State> >& successors,
            const State& state,
            const Action& action);
    void get_applicable_actions_sorted(
            std::vector<Action>& applicable,
            const State& state);

    const bool c_forward_updates;
    const bool c_backward_updates;
    const bool c_stop_at_inconsistent;
    const bool c_complete_policy_exploration;
    std::shared_ptr<heuristics::Heuristic<Reward> > m_heuristic;

    storage::SegmentedVector<PerStateInfo> m_state_infos;
};

template<typename Model, typename Reward>
HeuristicDepthFirstSearch<Model, Reward>::HeuristicDepthFirstSearch(
        std::shared_ptr<Model> t_model,
        std::shared_ptr<Reward> r_model,
        std::shared_ptr<heuristics::Heuristic<Reward> > heuristic,
        bool forward_updates,
        bool backward_updates,
        bool stop_at_inconsistent,
        bool complete_policy_exploration)
    : Engine<Model, Reward>(t_model, r_model)
    , c_forward_updates(forward_updates)
    , c_backward_updates(backward_updates)
    , c_stop_at_inconsistent(stop_at_inconsistent)
    , c_complete_policy_exploration(complete_policy_exploration)
    , m_heuristic(heuristic)
{}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::reset_and_prepare()
{
    for (int i = m_state_infos.size() - 1; i >= 0; i--) {
        PerStateInfo& info = m_state_infos[i];
        info.unset_solved();
        info.policy = UNDEFINED_ACTION;
    }
    prepare();
}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::prepare()
{
    storage::StateMap<SccInfo> scc_infos;
    std::deque<ExpansionInfo> expansion_queue;
    std::deque<State> stack;
    std::vector<Action> aops;

    State initial_state = this->m_transition_model->get_initial_state();
    // while (!initial_state_info.is_solved()) {
    unsigned iteration_num = 0;
    while (true) {
        assert(scc_infos.empty() && expansion_queue.empty() && stack.empty() && aops.empty());
        ++iteration_num;
        unsigned index = 0;
        bool last_flag = false;
        unsigned last_lowlink = 0;
        expansion_queue.emplace_back(initial_state);
        // std::cout << "i=" << iteration_num << " -> " << lookup_info(initial_state).reward << std::endl;
        while (!expansion_queue.empty()) {
            assert(aops.empty());
            ExpansionInfo& expand = expansion_queue.back();
            SccInfo& scc_info = scc_infos[expand.state];
            PerStateInfo& state_info = lookup_info(expand.state);
            if (scc_info.index == SccInfo::UNDEFINED) {
                //std::cout << "expand(" << expand.state << ")" << std::endl;
                scc_info.index = scc_info.lowlink = index++;
                last_lowlink = scc_info.lowlink;
                if (state_info.is_terminal() || state_info.is_solved()) {
                    last_flag = false;
                    state_info.set_solved();
                    expansion_queue.pop_back();
                    //std::cout << "  -> state#" << expand.state << " is terminal (or solved)" << std::endl;
                    continue;
                }
                if (c_forward_updates) {
                    expand.flag = !update_reward(
                            expand.successors, 
                            state_info,
                            expand.state).first;
                    if (state_info.is_terminal()) {
                        last_flag = expand.flag;
                        state_info.set_solved();
                        expansion_queue.pop_back();
                        continue;
                    }
                    if (c_stop_at_inconsistent && expand.flag) {
                        last_flag = true;
                        expansion_queue.pop_back();
                        continue;
                    }
                } else {
                    if (state_info.policy == UNDEFINED_ACTION) {
                        get_applicable_actions_sorted(aops, expand.state);
                        if (aops.empty()) {
                            last_flag = value_equal(state_info.reward, this->m_reward_model->get_min_reward());
                            state_info.reward = this->m_reward_model->get_min_reward();
                            state_info.set_terminal();
                            state_info.set_solved();
                            expansion_queue.pop_back();
                            continue;
                        }
                        state_info.policy = aops.front();
                        aops.clear();
                    }
                    this->m_transition_model->get_successors(
                        expand.successors,
                        expand.state,
                        state_info.policy);
                }
                sort_successors_for_exploration(
                        expand.successors,
                        expand.state,
                        state_info.policy);
                scc_info.onstack = true;
                stack.push_front(expand.state);
                //std::cout << "  -> pi(state#" << expand.state << ") = " << state_info.policy << " -> " << expand.successors.size() << " successors" << std::endl;
            } else {
                scc_info.lowlink = std::min(scc_info.lowlink, last_lowlink);
                expand.flag = expand.flag || last_flag;
                expand.successors.pop_back();
                if (!c_complete_policy_exploration && expand.flag) {
                    expand.successors.clear();
                }
            }
            while (!expand.successors.empty()) {
                State succ = expand.successors.back().second;
                SccInfo& succ_info = scc_infos[succ];
                if (succ_info.index == SccInfo::UNDEFINED) {
                    expansion_queue.emplace_back(succ);
                    break;
                } else if (succ_info.onstack) {
                    scc_info.lowlink = std::min(scc_info.lowlink, succ_info.index);
                } else if (c_complete_policy_exploration && !lookup_info(succ).is_solved()) {
                    expand.flag = true;
                }
                expand.successors.pop_back();
            }
            if (expand.successors.empty()) {
                //std::cout << "backtrack(" << expand.state << ") -> flag=" << expand.flag << std::endl;
                if (c_backward_updates && expand.flag) {
                    std::pair<bool, bool> update_info = update_reward(
                            expand.successors,
                            state_info,
                            expand.state);
                    expand.flag = expand.flag || !update_info.first || !update_info.second;
                }
                if (scc_info.index == scc_info.lowlink) {
                    // for (auto it = stack.begin();; it++) {
                    //     expand.flag = expand.flag || !update_reward(*it);
                    //     if (*it == expand.state) break;
                    // }
                    //std::cout << "scc [";
                    assert(!stack.empty());
                    auto end = stack.begin();
                    if (expand.flag) {
                        do {
                            //std::cout << (end != stack.begin() ? ", " : "") << (*end);
                            scc_infos[*end].onstack = false;
                            assert(*end == expand.state || (end+1) != stack.end());
                        } while (*(end++) != expand.state);
                    } else {
                        do {
                            //std::cout << (end != stack.begin() ? ", " : "") << (*end);
                            scc_infos[*end].onstack = false;
                            lookup_info(*end).set_solved();
                            assert(*end == expand.state || (end+1) != stack.end());
                        } while (*(end++) != expand.state);
                    }
                    //std::cout << "] completed" << std::endl;
                    stack.erase(stack.begin(), end);
                }
                last_flag = expand.flag;
                last_lowlink = scc_info.lowlink;
                expansion_queue.pop_back();
            }
        }
        assert(stack.empty());
        scc_infos.clear();
        if (!last_flag) {
            break;
        }
    }
}

template<typename Model, typename Reward>
ValueType 
HeuristicDepthFirstSearch<Model, Reward>::get_reward(const State& state)
{
    return lookup_info(state).reward;
}

template<typename Model, typename Reward>
Action 
HeuristicDepthFirstSearch<Model, Reward>::get_policy_action(const State& state)
{
    PerStateInfo& info = lookup_info(state);
    if (info.is_terminal()) {
        return UNDEFINED_ACTION;
    } else if (info.is_solved()) {
        return info.policy;
    } else {
        std::vector<Action> appl;
        get_applicable_actions_sorted(appl, state);
        if (appl.empty()) {
            info.reward = this->m_reward_model->get_min_reward();
            info.set_terminal();
            return UNDEFINED_ACTION;
        }
        return appl.front();
    }
}

template<typename Model, typename Reward>
typename HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo&
HeuristicDepthFirstSearch<Model, Reward>::lookup_info(const State& state)
{
    if (state >= m_state_infos.size()) {
        m_state_infos.resize(state + 1);
    }
    PerStateInfo& info = m_state_infos[state];
    if (info.is_new()) {
        info.reward = this->m_reward_model->get_max_reward();
        info.set_opened();
        if (this->m_reward_model->is_terminal(state)) {
            info.set_terminal();
        } else if (m_heuristic != nullptr) {
            heuristics::EvaluationResult est = m_heuristic->evaluate(state);
            if (est.dead) {
                info.reward = this->m_reward_model->get_min_reward();
                info.set_terminal();
            } else {
                info.reward = est.value;
            }
        }
    }
    return info;
}

template<typename Model, typename Reward>
std::pair<bool, bool>
HeuristicDepthFirstSearch<Model, Reward>::update_reward(
        std::vector<std::pair<ValueType, State> >& policy_successors,
        PerStateInfo& info,
        const State& state)
{
    assert(!info.is_terminal()); // && !info.is_solved());
    static std::vector<Action> applicable;
    static std::vector<std::pair<ValueType, State> > successors;
    assert(successors.empty() && applicable.empty());
    get_applicable_actions_sorted(applicable, state);
    if (applicable.empty()) {
        bool res = value_equal(info.reward, this->m_reward_model->get_min_reward());
        info.reward = this->m_reward_model->get_min_reward();
        info.set_terminal();
        return std::pair<bool, bool>(res, true);
    }
    std::sort(applicable.begin(), applicable.end());
    std::reverse(applicable.begin(), applicable.end());
    ValueType immediate = this->m_reward_model->get_immediate_reward(state);
    ValueType reward = this->m_reward_model->get_min_reward();
    Action policy = UNDEFINED_ACTION;
    for (int i = applicable.size() - 1; i >= 0; i--) {
        this->m_transition_model->get_successors(successors, state, applicable[i]);
        ValueType value = immediate;
        for (int j = successors.size() - 1; j >= 0; j--) {
            value += (successors[j].first *
                    (get_reward(successors[j].second) + this->m_reward_model->get_transition_reward(state, applicable[i], successors[j].second)));
        }
        if (value_greater(value, reward)
                || (applicable[i] == info.policy && value_equal(value, reward))) {
            reward = value;
            policy = applicable[i];
            policy_successors.swap(successors);
        }
        successors.clear();
    }
    applicable.clear();
    std::pair<bool, bool> res(
            value_equal(reward, info.reward),
            info.policy == UNDEFINED_ACTION || info.policy == policy);
    info.reward = reward;
    info.policy = policy;
    return res;
}

template<typename Model, typename Reward>
bool
HeuristicDepthFirstSearch<Model, Reward>::update_reward_get_policy(Action& p, const State& state)
{
    static std::vector<std::pair<ValueType, State> > aux;
    PerStateInfo& info = lookup_info(state);
    bool res = update_reward(aux, lookup_info(state), state).first;
    aux.clear();
    p = info.policy;
    return res;
}

template<typename Model, typename Reward>
void
HeuristicDepthFirstSearch<Model, Reward>::sort_successors_for_exploration(
        std::vector<std::pair<ValueType, State> >&,
        const State&,
        const Action&)
{
    // utils::g_rng.shuffle(successors);
}

template<typename Model, typename Reward>
void
HeuristicDepthFirstSearch<Model, Reward>::get_applicable_actions_sorted(
        std::vector<Action>& applicable,
        const State& state)
{
    static auto less = [](const Action& x, const Action& y) {
        return x > y;
    };
    this->m_transition_model->get_applicable_actions(applicable, state);
    std::sort(applicable.begin(), applicable.end(), less);
}

template<typename Model, typename Reward>
bool
HeuristicDepthFirstSearch<Model, Reward>::update_reward(const State& state)
{
    Action p; 
    return update_reward_get_policy(p, state);
}

template<typename Model, typename Reward>
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::PerStateInfo()
    : reward(0)
    , policy(UNDEFINED_ACTION)
    , status(0)
{
}

template<typename Model, typename Reward>
bool 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::is_new() const
{
    return !(status & OPENED);
}

template<typename Model, typename Reward>
bool 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::is_solved() const
{
    return (status & SOLVED);
}

template<typename Model, typename Reward>
bool 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::is_terminal() const
{
    return (status & TERMINAL);
}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::set_opened()
{
    assert(is_new() && !is_solved() && !is_terminal());
    status = status | OPENED;
    assert(!is_new() && !is_solved() && !is_terminal());
}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::set_solved()
{
    status = status | SOLVED;
    assert(is_solved());
}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::set_terminal()
{
    assert(!is_terminal());
    status = status | TERMINAL;
    assert(is_terminal());
}

template<typename Model, typename Reward>
void 
HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::unset_solved()
{
    status = status & ~SOLVED;
    assert(!is_solved());
}

template<typename Model, typename Reward>
const unsigned HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::OPENED   = 1;

template<typename Model, typename Reward>
const unsigned HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::SOLVED   = 2;

template<typename Model, typename Reward>
const unsigned HeuristicDepthFirstSearch<Model, Reward>::PerStateInfo::TERMINAL = 4;

template<typename Model, typename Reward>
const unsigned HeuristicDepthFirstSearch<Model, Reward>::SccInfo::UNDEFINED = -1;

}
}
}

#endif
