#include "hmax_dead_end_heuristic.h"
#include "../../sas_task/sas_task.h"
#include "../../sas_task/fact_pair.h"
#include "../../sas_task/action.h"
#include "../../storage/hash.h"
#include "../../cli/macros.h"

#include <unordered_set>
#include <algorithm>
#include <cassert>
#include <cstring>

namespace fd {
namespace heuristics {
namespace classic {

struct CounterHasher {
    CounterHasher(const std::vector<std::vector<unsigned> >& pres)
        : m_pres(pres)
    {}
    std::size_t operator()(const unsigned& c) const
    {
        storage::hash::HashState h;
        const auto& pre = m_pres[c];
        for (unsigned i = 0; i < pre.size(); i++) {
            h.feed(pre[i]);
        }
        return h.get_hash64();
    }
    const std::vector<std::vector<unsigned> >& m_pres;
};

struct CounterEqual {
    CounterEqual(const std::vector<std::vector<unsigned> >& pres)
        : m_pres(pres)
    {}
    bool operator()(const unsigned& c1, const unsigned& c2) const
    {
        const auto& pre1 = m_pres[c1];
        const auto& pre2 = m_pres[c2];
        if (pre1.size() != pre2.size()) {
            return false;
        }
        for (unsigned i = 0; i < pre1.size(); i++) {
            if (pre1[i] != pre2[i]) {
                return false;
            }
        }
        return true;
    }
    const std::vector<std::vector<unsigned> >& m_pres;
};

HmaxDeadEndHeuristic::HmaxDeadEndHeuristic(
        const std::shared_ptr<sas_task::SASTask> task)
    : ClassicalPlanningHeuristic(task)
{
    FD_LOG_MESSAGE_T("Preparing hmax dead end detection heuristic...");
    if (task->has_axioms() || task->has_conditional_effects()) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::UNSUPPORTED_FEATURE, "hmax does currently not support axioms and conditional effects");
    }
    m_num_facts = 0;
    m_fact_map.resize(task->get_num_variables());
    for (unsigned var = 0; var < m_fact_map.size(); var++) {
        m_fact_map[var].resize(task->get_variable_domain_size(var));
        for (unsigned val = 0; val < task->get_variable_domain_size(var); val++) {
            m_fact_map[var][val] = m_num_facts++;
        }
    }

    m_num_goals = task->get_num_goals();
    m_in_goal.resize(m_num_facts+1, false);
    for (unsigned i = 0; i < task->get_num_goals(); i++) {
        const sas_task::FactPair& p = task->get_goal_fact(i);
        m_in_goal[m_fact_map[p.var][p.value]] = true;
    }

    m_counter_preconditions.resize(m_num_facts + 1);
    m_counter_effects.resize(task->get_num_actions());
    m_num_preconditions.resize(task->get_num_actions(), 0);

    std::vector<std::vector<Fact> > preconditions(task->get_num_actions());
    std::unordered_set<Counter, CounterHasher, CounterEqual> counter_set(
            task->get_num_actions(),
            CounterHasher(preconditions),
            CounterEqual(preconditions));
    Counter c = 0;
    for (unsigned i = 0; i < task->get_num_actions(); i++) {
        const sas_task::Action& action = task->get_action(i);
        std::vector<Fact>& pre = preconditions[c];
        pre.resize(action.get_preconditions().size());
        for (unsigned j = 0; j < pre.size(); j++) {
            const sas_task::FactPair& p = action.get_preconditions()[j];
            pre[j] = m_fact_map[p.var][p.value];
        }
        std::sort(pre.begin(), pre.end());
        pre.erase(std::unique(pre.begin(), pre.end()), pre.end());
        auto x = counter_set.insert(c);
        if (!x.second) {
            pre.clear();
        } else {
            if (!pre.empty()) {
                for (const auto& f : pre) {
                    m_counter_preconditions[f].push_back(c);
                }
                m_num_preconditions[c] = pre.size();
            } else {
                m_counter_preconditions.back().push_back(c);
                m_num_preconditions[c] = 1;
            }
            c++;
        }
        for (unsigned j = 0; j < action.get_effects().size(); j++) {
            const sas_task::Effect& eff = action.get_effects()[j];
            const sas_task::FactPair& p = eff.get_effect();
            m_counter_effects[*x.first].push_back(m_fact_map[p.var][p.value]);
        }
    }

    m_counter_effects.resize(c);
    m_num_preconditions.resize(c);
    for (unsigned i = 0; i < c; i++) {
        std::vector<Fact>& eff = m_counter_effects[i];
        std::sort(eff.begin(), eff.end());
        eff.erase(std::unique(eff.begin(), eff.end()), eff.end());
    }

    m_num_facts++;

    FD_LOG_MESSAGE_T("hmax initialization done!");
}

EvaluationResult
HmaxDeadEndHeuristic::evaluate(const StateData& state)
{
    static std::vector<bool> reached(m_num_facts);
    static std::vector<Fact> queue(m_num_facts);
    static std::vector<int>  pres(m_num_preconditions.size());
    memcpy(pres.data(), m_num_preconditions.data(), m_num_preconditions.size() * sizeof(int));
    std::fill(reached.begin(), reached.end(), false);
    queue.clear();
    unsigned open_goals = m_num_goals;
    for (unsigned var = 0; var < m_task->get_num_variables(); var++) {
        Fact f = m_fact_map[var][state[var]];
        reached[f] = true;
        queue.push_back(f);
        if (m_in_goal[f]) {
            open_goals--;
        }
    }
    reached.back() = true;
    queue.push_back(m_num_facts - 1);
    while (!queue.empty() && open_goals > 0) {
        Fact f = queue.back();
        queue.pop_back();
        const std::vector<Counter>& in_pre = m_counter_preconditions[f];
        for (int i = in_pre.size() - 1; i >= 0; i--) {
            Counter c = in_pre[i];
            if (--pres[c] == 0) {
                const std::vector<Fact>& eff = m_counter_effects[c];
                for (int j = eff.size() - 1; j >= 0; j--) {
                    Fact fprime = eff[j];
                    if (!reached[fprime]) {
                        reached[fprime] = true;
                        if (m_in_goal[fprime]) {
                            open_goals--;
                        }
                        queue.push_back(fprime);
                    }
                }
            }
        }
    }
    return EvaluationResult(open_goals != 0, 0);
}

}
}
}
