#include "probabilistic_sas_task.h"
#include "sas_task.h"
#include "../types.h"
#include "../cli/macros.h"

#include <cassert>
#include <map>
#include <sstream>
#include <algorithm>

namespace fd {
namespace sas_task {

struct OutcomeData
{
    int numerator;
    int denominator;
    std::string action_name;
    OutcomeData(int num, int den, const std::string& name)
        : numerator(num), denominator(den), action_name(name) {}
};

static OutcomeData
parse_action_outcome_name(const std::string& name)
{
    std::size_t det_pos = name.find("_DETDUP_");
    if (det_pos != std::string::npos) {
        std::size_t weight_pos = name.find("_WEIGHT_");
        if (weight_pos != std::string::npos) {
            std::size_t weight_num_pos = weight_pos + 8;
            std::string tmp = name.substr(weight_num_pos, std::string::npos);
            weight_num_pos = tmp.find("_");
            double numerator, denominator;
            std::istringstream(tmp.substr(0, weight_num_pos)) >> numerator;
            std::istringstream(tmp.substr(weight_num_pos + 1,
                                          std::string::npos)) >> denominator;
            return OutcomeData((int)numerator,
                               (int)denominator,
                               name.substr(0, det_pos)
                               + name.substr(name.find(" "), std::string::npos));
        } else {
            return OutcomeData(1, 1, name.substr(0, det_pos)
                               + name.substr(name.find(" "), std::string::npos));
        }
    } else {
        return OutcomeData(1, 1, name);
    }
}

static void
get_precondition(std::vector<FactPair>& prec,
                 const SASTask& task,
                 int index,
                 bool is_axiom)
{
    const std::vector<FactPair>* apre = NULL;
    if (is_axiom) {
        apre = &task.get_axiom(index).get_preconditions();
    } else {
        apre = &task.get_action(index).get_preconditions();
    }
    prec.insert(prec.end(), apre->begin(), apre->end());
    std::sort(prec.begin(), prec.end());
}

static void
get_effects(std::vector<Effect>& eff,
            const SASTask& task,
            int index,
            bool is_axiom)
{
    const std::vector<Effect>* aeff = NULL;
    if (is_axiom) {
        aeff = &task.get_axiom(index).get_effects();
    } else {
        aeff = &task.get_action(index).get_effects();
    }
    eff.insert(eff.end(), aeff->begin(), aeff->end());
}

static int
gcd(int a, int b)
{
    for (;;) {
        if (a == 0) {
            return b;
        }
        b %= a;
        if (b == 0) {
            return a;
        }
        a %= b;
    }
}

static int
lcm(int a, int b)
{
    int temp = gcd(a, b);

    return temp ? (a / temp * b) : 0;
}

ProbabilisticSASTask::ProbabilisticSASTask(const std::shared_ptr<SASTask> task)
    : m_sas_task(task)
{
    reconstruct_probabilistic_actions(*task);
}

ProbabilisticSASTask::ProbabilisticSASTask(std::istream& in)
{
    m_sas_task = read_sas_task(in);
    reconstruct_probabilistic_actions(*m_sas_task);
}

void
ProbabilisticSASTask::reconstruct_probabilistic_actions(const SASTask& task)
{
    FD_LOG_MESSAGE_T("reconstructing probabilistic actions...");
    // int num_vars = task.get_num_variables();
    // m_variable_domain.resize(num_vars);
    // m_axiom_layers.resize(num_vars);
    // m_default_axiom_values.resize(num_vars);
    // m_variable_names.resize(num_vars);
    // const std::vector<int>& iv = task.get_initial_state_values();
    // m_initial_state_values.insert(m_initial_state_values.end(),
    //                                       iv.begin(),
    //                                       iv.end());
    // for (int var = num_vars - 1; var >= 0; var--) {
    //     m_variable_domain[var] = task.get_variable_domain_size(var);
    //     m_axiom_layers[var] = task.get_variable_axiom_layer(var);
    //     m_default_axiom_values[var] = task.get_variable_default_axiom_value(
    //             var);
    //     m_variable_names[var] = task.get_variable_name(var);
    // }

    // m_fact_names.resize(num_vars);
    // m_inconsistent_facts.resize(num_vars);
    // for (int var = num_vars - 1; var >= 0; var--) {
    //     auto& names = m_fact_names[var];
    //     names.resize(m_variable_domain[var]);
    //     auto& inco = m_inconsistent_facts[var];
    //     inco.resize(m_variable_domain[var]);
    //     for (int val = m_variable_domain[var] - 1; val >= 0; val--) {
    //         names[val] = task.get_fact_name(FactPair(var, val));
    //         FactPair p(var, val);
    //         const std::set<FactPair>& facts = task.get_mutexes_for_fact(p);
    //         inco[val].insert(facts.begin(), facts.end());
    //     }
    // }

    // m_goal.reserve(task.get_num_goals());
    // for (int i = 0; i < task.get_num_goals(); i++) {
    //     m_goal.push_back(task.get_goal_fact(i));
    // }

    // m_axioms.reserve(task.get_num_axioms());
    // for (int axiom = 0; axiom < task.get_num_axioms(); axiom++) {
    //     m_axioms.emplace_back(task.get_axiom(axiom));
    // }

    // reconstructing probabilistic actions from outcomes
    std::map<std::string, unsigned> name_to_i;
    std::vector<std::vector<std::pair<int, int> > > probabilities;
    unsigned cur_i = 0;
    for (unsigned orig_op = 0; orig_op < task.get_num_actions(); orig_op++) {
        OutcomeData data =
            parse_action_outcome_name(task.get_action(orig_op).get_name());
        if (data.numerator == 0) {
            continue;
        }
        unsigned i;
        auto it = name_to_i.find(data.action_name);
        if (it == name_to_i.end()) {
            i = cur_i++;
            name_to_i[data.action_name] = i;
            probabilities.emplace_back();
            m_prob_actions.emplace_back(data.action_name, i);
            ProbabilisticAction& op = m_prob_actions.back();
            get_precondition(op.get_preconditions(), task, orig_op, false);
        } else {
            i = it->second;
#ifndef NDEBUG
            std::vector<FactPair> pre_old(m_prob_actions[i].get_preconditions());
            std::vector<FactPair> pre_new;
            get_precondition(pre_new, task, orig_op, false);
            assert(pre_old.size() == pre_new.size());
            std::sort(pre_old.begin(), pre_old.end());
            std::sort(pre_new.begin(), pre_new.end());
            for (unsigned j = 0; j < pre_old.size(); j++) {
                for (unsigned k = 0; k < pre_old.size(); k++) {
                    assert(pre_old[k] == pre_new[k]);
                }
            }
#endif
        }
        probabilities[i].emplace_back(data.numerator, data.denominator);
        ProbabilisticAction& op = m_prob_actions[i];
        op.get_outcomes().emplace_back();
        get_effects(op.get_outcomes().back().get_effects(), task, orig_op, false);
    }

    // computing probabilities
    for (unsigned i = 0; i < m_prob_actions.size(); i++) {
        ProbabilisticAction& op = m_prob_actions[i];
        int total = 1;
        int sum = 0;
        for (const auto& prob : probabilities[i]) {
            total = lcm(total, prob.second);
        }
        for (unsigned j = 0; j < probabilities[i].size(); j++) {
            int x = total / probabilities[i][j].second;
            int n = x * probabilities[i][j].first;
            op.get_outcomes()[j].set_probability(fraction_to_value(n, total));
            sum += n;
        }
        assert(sum <= total);
        if (sum < total) {
            op.get_outcomes().emplace_back();
            op.get_outcomes().back().set_probability(fraction_to_value(total - sum, total));
        }
        op.normalize();
    }

    unsigned num_outcomes = 0;
    for (unsigned i = 0; i < m_prob_actions.size(); i++) {
        num_outcomes += m_prob_actions[i].get_outcomes().size();
    }

    // m_has_ce = task.has_conditional_effects();
    FD_LOG_MESSAGE_T("done reconstructing probabilistic actions!");
    FD_LOG_MESSAGE("probabilistic actions: " << get_num_actions() << " (with " << num_outcomes << " outcomes)");
}


unsigned
ProbabilisticSASTask::get_num_variables() const
{
    return m_sas_task->get_num_variables();
}

std::string
ProbabilisticSASTask::get_variable_name(int var) const
{
    return m_sas_task->get_variable_name(var);
}

unsigned
ProbabilisticSASTask::get_variable_domain_size(int var) const
{
    return m_sas_task->get_variable_domain_size(var);
}

int
ProbabilisticSASTask::get_variable_axiom_layer(int var) const
{
    return m_sas_task->get_variable_axiom_layer(var);
}

int
ProbabilisticSASTask::get_variable_default_axiom_value(int var) const
{
    return m_sas_task->get_variable_default_axiom_value(var);
}

std::string
ProbabilisticSASTask::get_fact_name(const FactPair& fact) const
{
    return m_sas_task->get_fact_name(fact);
}

bool
ProbabilisticSASTask::are_facts_mutex(const FactPair& fact1, const FactPair& fact2) const
{
    return m_sas_task->are_facts_mutex(fact1, fact2);
}

const std::vector<int>&
ProbabilisticSASTask::get_initial_state_values() const
{
    return m_sas_task->get_initial_state_values();
}

unsigned
ProbabilisticSASTask::get_num_goals() const
{
    return m_sas_task->get_num_goals();
}

const FactPair&
ProbabilisticSASTask::get_goal_fact(int x) const
{
    return m_sas_task->get_goal_fact(x);
}

unsigned
ProbabilisticSASTask::get_num_actions() const
{
    return m_prob_actions.size();
}

unsigned
ProbabilisticSASTask::get_num_axioms() const
{
    return m_sas_task->get_num_axioms();
}

const ProbabilisticAction&
ProbabilisticSASTask::get_action(int a) const
{
    return m_prob_actions[a];
}

const Action&
ProbabilisticSASTask::get_axiom(int a) const
{
    return m_sas_task->get_axiom(a);
}

bool
ProbabilisticSASTask::has_axioms() const
{
    return m_sas_task->has_axioms();
}

bool
ProbabilisticSASTask::has_conditional_effects() const
{
    return m_sas_task->has_conditional_effects();
}

const std::set<FactPair>&
ProbabilisticSASTask::get_mutexes_for_fact(const FactPair& f) const
{
    return m_sas_task->get_mutexes_for_fact(f);
}

const std::shared_ptr<SASTask>
ProbabilisticSASTask::get_determinization() const
{
    return m_sas_task;
}

}
}
