#include "sas_task.h"
#include "sas_parser_utils.h"
#include "../cli/macros.h"

#include <algorithm>
#include <cassert>
#include <set>
#include <vector>

using namespace std;

namespace fd {
namespace sas_task {

void
read_and_verify_version(istream& in)
{
    int version;
    check_magic(in, "begin_version");
    in >> version;
    check_magic(in, "end_version");
    if (version != PRE_FILE_VERSION) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::INPUT_ERROR,
                "Expected translator output file version " << PRE_FILE_VERSION
             << ", got " << version << "." << endl
             << "Exiting.");
    }
}

bool
read_metric(istream& in)
{
    bool use_metric;
    check_magic(in, "begin_metric");
    in >> use_metric;
    check_magic(in, "end_metric");
    return use_metric;
}

vector<Variable>
read_variables(istream& in)
{
    int count;
    in >> count;
    vector<Variable> variables;
    variables.reserve(count);
    for (int i = 0; i < count; ++i) {
        variables.emplace_back(in);
    }
    return variables;
}

vector<vector<set<FactPair>>>
read_mutexes(istream& in, const vector<Variable>& variables)
{
    vector<vector<set<FactPair>>> inconsistent_facts(variables.size());
    for (size_t i = 0; i < variables.size(); ++i) {
        inconsistent_facts[i].resize(variables[i].domain_size);
    }

    int num_mutex_groups;
    in >> num_mutex_groups;

    /*
      NOTE: Mutex groups can overlap, in which case the same mutex
      should not be represented multiple times. The current
      representation takes care of that automatically by using sets.
      If we ever change this representation, this is something to be
      aware of.
    */
    for (int i = 0; i < num_mutex_groups; ++i) {
        check_magic(in, "begin_mutex_group");
        int num_facts;
        in >> num_facts;
        vector<FactPair> invariant_group;
        invariant_group.reserve(num_facts);
        for (int j = 0; j < num_facts; ++j) {
            int var;
            int value;
            in >> var >> value;
            invariant_group.emplace_back(var, value);
        }
        check_magic(in, "end_mutex_group");
        for (const FactPair& fact1 : invariant_group) {
            for (const FactPair& fact2 : invariant_group) {
                if (fact1.var != fact2.var) {
                    /* The "different variable" test makes sure we
                       don't mark a fact as mutex with itself
                       (important for correctness) and don't include
                       redundant mutexes (important to conserve
                       memory). Note that the translator (at least
                       with default settings) removes mutex groups
                       that contain *only* redundant mutexes, but it
                       can of course generate mutex groups which lead
                       to *some* redundant mutexes, where some but not
                       all facts talk about the same variable. */
                    inconsistent_facts[fact1.var][fact1.value].insert(fact2);
                }
            }
        }
    }
    return inconsistent_facts;
}

vector<FactPair>
read_goal(istream& in)
{
    check_magic(in, "begin_goal");
    vector<FactPair> goals = read_facts(in);
    check_magic(in, "end_goal");
    if (goals.empty()) {
        FD_LOG_ERROR_AND_EXIT(cli::ExitCode::INPUT_ERROR, "Task has no goal condition!");
    }
    return goals;
}

vector<Action>
read_actions(
    istream& in, bool is_axiom, bool use_metric,
    const vector<Variable>& variables)
{
    int count;
    in >> count;
    vector<Action> actions;
    actions.reserve(count);
    for (int i = 0; i < count; ++i) {
        actions.emplace_back(in, is_axiom, use_metric, i);
        check_facts(actions.back(), variables);
    }
    return actions;
}

SASTask::SASTask(std::istream& in)
    : has_ce(false)
{
    FD_LOG_MESSAGE_T("reading input...");
    read_and_verify_version(in);
    bool use_metric = read_metric(in);
    variables = read_variables(in);
    int num_variables = variables.size();

    mutexes = read_mutexes(in, variables);

    initial_state_values.resize(num_variables);
    check_magic(in, "begin_state");
    for (int i = 0; i < num_variables; ++i) {
        in >> initial_state_values[i];
    }
    check_magic(in, "end_state");

    for (int i = 0; i < num_variables; ++i) {
        variables[i].axiom_default_value = initial_state_values[i];
    }

    goals = read_goal(in);
    check_facts(goals, variables);
    actions = read_actions(in, false, use_metric, variables);
    axioms = read_actions(in, true, use_metric, variables);
    /* TODO: We should be stricter here and verify that we
       have reached the end of "in". */

    for (unsigned i = 0; !has_ce && i < actions.size(); i++) {
        for (unsigned j = 0; !has_ce && j < actions[i].get_effects().size(); j++) {
            has_ce = actions[i].get_effects()[j].is_conditional_effect();
        }
    }

    FD_LOG_MESSAGE_T("done reading input!");
    FD_LOG_MESSAGE("variables: " << get_num_variables());
    FD_LOG_MESSAGE("goals: " << get_num_goals());
    FD_LOG_MESSAGE("actions: " << get_num_actions());
    FD_LOG_MESSAGE("axioms: " << get_num_axioms());
    FD_LOG_MESSAGE("has conditional effects: " << has_conditional_effects());
}

unsigned
SASTask::get_num_variables() const
{
    return variables.size();
}

string
SASTask::get_variable_name(int var) const
{
    return variables[var].name;
}

unsigned
SASTask::get_variable_domain_size(int var) const
{
    return variables[var].domain_size;
}

int
SASTask::get_variable_axiom_layer(int var) const
{
    return variables[var].axiom_layer;
}

int
SASTask::get_variable_default_axiom_value(int var) const
{
    return variables[var].axiom_default_value;
}

string
SASTask::get_fact_name(const FactPair& fact) const
{
    return variables[fact.var].fact_names[fact.value];
}

bool
SASTask::are_facts_mutex(const FactPair& fact1, const FactPair& fact2) const
{
    if (fact1.var == fact2.var) {
        // Same variable: mutex iff different value.
        return fact1.value != fact2.value;
    }
    return bool(mutexes[fact1.var][fact1.value].count(fact2));
}

const vector<int>&
SASTask::get_initial_state_values() const
{
    return initial_state_values;
}

unsigned
SASTask::get_num_goals() const
{
    return goals.size();
}

const FactPair&
SASTask::get_goal_fact(int index) const
{
    return goals[index];
}

unsigned
SASTask::get_num_actions() const
{
    return actions.size();
}

unsigned
SASTask::get_num_axioms() const
{
    return axioms.size();
}

const Action&
SASTask::get_action(int x) const
{
    assert(((unsigned) x) < actions.size());
    return actions[x];
}

const Action&
SASTask::get_axiom(int x) const
{
    assert(((unsigned) x) < axioms.size());
    return axioms[x];
}

bool
SASTask::has_axioms() const
{
    return get_num_axioms() > 0;
}

bool
SASTask::has_conditional_effects() const
{
    return has_ce;
}

const std::set<FactPair>&
SASTask::get_mutexes_for_fact(const FactPair& f) const
{
    return mutexes[f.var][f.value];
}

std::shared_ptr<SASTask>
read_sas_task(std::istream& in)
{
    return make_shared<SASTask>(in);
}

}
}
