#include "array_compressor.h"

#include <cassert>
#include <algorithm>
#include <deque>

namespace fd {
namespace storage {
namespace compression {

const unsigned ArrayCompressor::BinInfo::UNDEFINED = -1;

ArrayCompressor::ArrayCompressor(const std::vector<int>& upper_bounds)
    : ArrayCompressor(
            std::vector<int>(upper_bounds.size(), 0),
            upper_bounds)
{}

ArrayCompressor::ArrayCompressor(
        const std::vector<int>& lower_bounds,
        const std::vector<int>& upper_bounds)
    : m_num_bins(0)
{
    assert(lower_bounds.size() == upper_bounds.size());
    m_bin_infos.resize(lower_bounds.size());

    unsigned bin_size = sizeof(Bin) * 8;

    std::vector<std::vector<unsigned> > bits_to_vars(bin_size + 1);
    for (unsigned i = 0; i < lower_bounds.size(); i++) {
        assert(lower_bounds[i] < upper_bounds[i]);
        unsigned values = ((unsigned) upper_bounds[i]) - ((unsigned) lower_bounds[i]);

        int msb = 0;
        while ((1U << msb) < (size_t) values) {
            ++msb;
        }

        assert((unsigned)msb <= bin_size);
        bits_to_vars[msb].push_back(i);
        m_bin_infos[i].lb = lower_bounds[i];
    }
    
    Bin one = 1;
    int vars_left = lower_bounds.size();
    while (vars_left > 0) {
        m_num_bins++;
        unsigned bits_left = bin_size;
        while (bits_left > 0) {
            unsigned b = bits_left;
            while (b > 0 && bits_to_vars[b].empty()) {
                b--;
            }
            if (b == 0) {
                break;
            }
            unsigned i = bits_to_vars[b].back();
            bits_to_vars[b].pop_back();
            BinInfo& info = m_bin_infos[i];
            info.bin_id = m_num_bins - 1;
            if (b == bin_size) {
                info.mask = -1;
                info.shift = 0;
            } else {
                info.mask  = ((one << b) - 1) << (bin_size - bits_left);
                info.shift = (bin_size - bits_left);
            }
            bits_left -= b;
            vars_left--;
        }
    }
}

int
ArrayCompressor::get(const Bin* bins, unsigned var) const
{
    assert(var < m_bin_infos.size());
    const BinInfo& info = m_bin_infos[var];
    return ((int) ((bins[info.bin_id] & info.mask) >> info.shift)) + info.lb;
}

void
ArrayCompressor::set(Bin* bins, unsigned var, int val) const
{
    assert(var < m_bin_infos.size());
    assert(val >= m_bin_infos[var].lb);
    const BinInfo& info = m_bin_infos[var];
    Bin& bin = bins[info.bin_id];
    bin = (bin & ~info.mask) | (((Bin) (val - info.lb)) << info.shift);
}

unsigned
ArrayCompressor::get_num_bins() const
{
    return m_num_bins;
}



CompressedArray::CompressedArray(
        const ArrayCompressor& compressor)
    : m_array(new Bin[compressor.get_num_bins()])
    , m_compressor(&compressor)
{}

CompressedArray::CompressedArray(
        CompressedArray&& other)
    : m_array(other.m_array)
    , m_compressor(other.m_compressor)
{
    other.m_array = NULL;
}

CompressedArray::~CompressedArray()
{
    if (m_array) {
        delete[] m_array;
    }
}

int
CompressedArray::operator[](unsigned i) const
{
    return m_compressor->get(m_array, i);
}

int
CompressedArray::at(unsigned i) const
{
    return m_compressor->get(m_array, i);
}

void 
CompressedArray::set(unsigned x, int y)
{
    m_compressor->set(m_array, x, y);
}

void
CompressedArray::copy(Bin* dest) const
{
    memcpy(dest, m_array, m_compressor->get_num_bins() * sizeof(Bin));
}

Bin*
CompressedArray::data()
{
    return m_array;
}

const Bin*
CompressedArray::data() const
{
    return m_array;
}


CompressedArrayConstReference::CompressedArrayConstReference(
        const Bin* b,
        const ArrayCompressor& compressor)
    : m_array(b)
    , m_compressor(&compressor)
{}

int
CompressedArrayConstReference::operator[](unsigned i) const
{
    return m_compressor->get(m_array, i);
}

int
CompressedArrayConstReference::at(unsigned i) const
{
    return m_compressor->get(m_array, i);
}

void
CompressedArrayConstReference::copy(Bin* dest) const
{
    memcpy(dest, m_array, m_compressor->get_num_bins() * sizeof(Bin));
}

const Bin*
CompressedArrayConstReference::data() const
{
    return m_array;
}

}
}
}
