#ifndef ARRAY_COMPRESSOR_H
#define ARRAY_COMPRESSOR_H

#include <vector>
#include <cstring>

namespace fd {
namespace storage {
namespace compression {

using Bin = unsigned int;

class ArrayCompressor {
public:
    ArrayCompressor(const std::vector<int>& upper_bounds);
    ArrayCompressor(const std::vector<int>& lower_bounds,
               const std::vector<int>& upper_bounds);
    ArrayCompressor() = default;

    int get(const Bin* bins, unsigned var) const;
    void set(Bin* bins, unsigned var, int val) const;
    template<typename ValueArray>
    void pack(Bin* bins, const ValueArray& values) const;
    template<typename ValueArray>
    void unpack(ValueArray& values, const Bin* bins) const;
    unsigned get_num_bins() const;

private:
    struct BinInfo {
        static const unsigned UNDEFINED;
        int lb;
        unsigned bin_id;
        Bin mask;
        unsigned shift;
        BinInfo()
            : lb(0)
            , bin_id(UNDEFINED)
            , mask(0)
            , shift(0)
        {}
    };
    std::vector<BinInfo> m_bin_infos;
    unsigned m_num_bins;
};


class CompressedArrayConstReference {
public:
    CompressedArrayConstReference(
            const Bin* b,
            const ArrayCompressor& compressor);
    CompressedArrayConstReference(CompressedArrayConstReference&& other) = default;
    CompressedArrayConstReference(const CompressedArrayConstReference& other) = default;
    ~CompressedArrayConstReference() = default;
    int operator[](unsigned) const;
    int at(unsigned) const;
    template<typename ValueArray>
    void unpack(ValueArray& array) const;
    void copy(Bin* dest) const;
    const Bin* data() const;
private:
    const Bin* m_array;
    const ArrayCompressor* m_compressor;
};


class CompressedArray {
public:
    CompressedArray(const ArrayCompressor& compressor);
    CompressedArray(CompressedArray&& other);
    ~CompressedArray();
    int operator[](unsigned) const;
    int at(unsigned) const;
    void set(unsigned, int);
    template<typename ValueArray>
    void unpack(ValueArray& array) const;
    void copy(Bin* dest) const;
    const Bin* data() const;
    Bin* data();
private:
    Bin* m_array;
    const ArrayCompressor* m_compressor;
};


template<typename ValueArray>
void
ArrayCompressor::pack(Bin* bins, const ValueArray& values) const
{
    memset(bins, 0, m_num_bins * sizeof(Bin));
    for (int var = m_bin_infos.size() - 1; var >= 0; var--) {
        set(bins, var, values[var]);
    }
}

template<typename ValueArray>
void
ArrayCompressor::unpack(ValueArray& values, const Bin* bins) const
{
    for (int var = m_bin_infos.size() - 1; var >= 0; var--) {
        values[var] = get(bins, var);
    }
}

template<typename ValueArray>
void
CompressedArrayConstReference::unpack(ValueArray& array) const
{
    m_compressor->unpack(array, m_array);
}

template<typename ValueArray>
void
CompressedArray::unpack(ValueArray& array) const
{
    m_compressor->unpack(array, m_array);
}

}
}
}

#endif
