#ifndef CONJUNCTIVE_CONDITION_EVALUATOR_H
#define CONJUNCTIVE_CONDITION_EVALUATOR_H

#include <vector>
#include <algorithm>
#include <cassert>
#include <tuple>
#include <iostream>

namespace fd {
namespace storage {

template<typename V>
class ConjunctiveConditionEvaluator {
public:
    ConjunctiveConditionEvaluator() : m_var(-1) {}
    ConjunctiveConditionEvaluator(ConjunctiveConditionEvaluator&& other);
    ~ConjunctiveConditionEvaluator();
    ConjunctiveConditionEvaluator& operator=(ConjunctiveConditionEvaluator&& other);

    template<typename ValueArray>
    void evaluate(std::vector<V>& result, const ValueArray& values) const;

    static ConjunctiveConditionEvaluator create(
            const std::vector<int>& domains,
            const std::vector<V>& values,
            const std::vector<std::vector<std::pair<int, int> > >& conds);

private:
    ConjunctiveConditionEvaluator(int var, unsigned size);

    struct SuccessorStore {
        SuccessorStore(unsigned size)
            : m_successors(size, NULL)
        {}

        SuccessorStore(SuccessorStore&& store)
            : m_successors(store.m_successors)
        {
            store.m_successors.clear();
        }

        ~SuccessorStore()
        {
            delete_objects();
        }

        SuccessorStore& operator=(SuccessorStore&& other)
        {
            delete_objects();
            m_successors = std::move(other.m_successors);
            other.m_successors.clear();
        }

        ConjunctiveConditionEvaluator<V>*&
        operator[](unsigned i)
        {
            assert(i < m_successors.size());
            return m_successors[i];
        }

        const ConjunctiveConditionEvaluator<V>*
        operator[](unsigned i) const
        {
            assert(i < m_successors.size());
            return m_successors[i];
        }

    private:
        void delete_objects()
        {
            for (unsigned i = 0; i < m_successors.size(); i++) {
                if (m_successors[i] != NULL) {
                    delete(m_successors[i]);
                }
            }
            m_successors.clear();
        }

        std::vector<ConjunctiveConditionEvaluator<V>* > m_successors;
    };

    int m_var;
    std::vector<V> m_immediate;
    ConjunctiveConditionEvaluator<V>* m_default;
    SuccessorStore m_successors;
};

}
}


namespace fd {
namespace storage {

template<typename V>
ConjunctiveConditionEvaluator<V>::ConjunctiveConditionEvaluator(
    ConjunctiveConditionEvaluator<V>&& other)
    : m_var(std::move(other.m_var))
    , m_default(other.m_default)
    , m_successors(std::move(other.m_successors))
{
    other.m_default = NULL;
}

template<typename V>
ConjunctiveConditionEvaluator<V>::ConjunctiveConditionEvaluator(
        int var,
        unsigned size)
    : m_var(var)
    , m_default(NULL)
    , m_successors(size)
{
}

template<typename V>
ConjunctiveConditionEvaluator<V>::~ConjunctiveConditionEvaluator()
{
    if (m_default != NULL) {
        delete(m_default);
    }
}

template<typename V>
ConjunctiveConditionEvaluator<V>&
ConjunctiveConditionEvaluator<V>::operator=(ConjunctiveConditionEvaluator<V>&& other)
{
    m_var = other.m_var;
    m_default = other.m_default;
    m_successors = std::move(other.m_successors);
    other.m_default = NULL;
    return *this;
}

template<typename V>
template<typename ValueArray>
void
ConjunctiveConditionEvaluator<V>::evaluate(
        std::vector<V>& result,
        const ValueArray& values) const
{
    result.insert(result.end(), m_immediate.begin(), m_immediate.end());

    if (m_default != NULL) {
        m_default->evaluate(result, values);
    }

    if (m_var >= 0) {
        const ConjunctiveConditionEvaluator<V>* succ = m_successors[values[m_var]];
        if (succ != NULL) {
            succ->evaluate(result, values);
        }
    }
}

template<typename V>
ConjunctiveConditionEvaluator<V> 
ConjunctiveConditionEvaluator<V>::create(
            const std::vector<int>& domains,
            const std::vector<V>& values,
            const std::vector<std::vector<std::pair<int, int> > >& conds)
{
    assert(values.size() == conds.size());

    std::vector<int> var_in_conditions(domains.size(), 0);
    std::vector<std::vector<int> > val_in_conditions(domains.size());
    for (unsigned var = 0; var < domains.size(); var++) {
        val_in_conditions[var].resize(domains[var], 0);
    }
    for (int i = conds.size() - 1; i >= 0; i--) {
        const std::vector<std::pair<int, int> >& cs = conds[i];
        for (const auto& c : cs) {
            var_in_conditions[c.first]++;
            val_in_conditions[c.first][c.second]++;
        }
    }

#if 1
    std::vector<std::tuple<int, int, int> > var_keys;
    // std::vector<int> var_keys;
    var_keys.reserve(domains.size());
    int min_var = 0;
    for (unsigned var = 0; var < domains.size(); var++) {
        // var_keys.push_back(var);
        std::sort(val_in_conditions[var].begin(), val_in_conditions[var].end());
        int l = (int) (0.15f * (float) domains[var]);
        int r = std::min((int) (0.85f * (float) domains[var]), (int)domains[var]-1);
        if (l == r) {
            l = 0;
            r = domains[var]-1;
        }
        var_keys.emplace_back(
                -var_in_conditions[var],
                -var_in_conditions[var]/domains[var],
                val_in_conditions[var][r] - val_in_conditions[var][l]);
        if (var_keys[min_var] > var_keys[var]) {
            min_var = var;
        }
    }
#else
    std::vector<int> var_keys;
    var_keys.reserve(domains.size());
    int min_var = 0;
    for (unsigned var = 0; var < domains.size(); var++) {
        var_keys.push_back(var);
    }
#endif
    std::vector<int>().swap(var_in_conditions);
    std::vector<std::vector<int> >().swap(val_in_conditions);

    ConjunctiveConditionEvaluator<V> result(min_var, domains[min_var]);
    std::vector<std::pair<int, int> > cond_sorted;
    for (unsigned i = 0; i < conds.size(); i++) {
        cond_sorted.insert(cond_sorted.end(), conds[i].begin(), conds[i].end());
        std::sort(cond_sorted.begin(), cond_sorted.end(), [&var_keys](
                    const std::pair<int, int>& p,
                    const std::pair<int, int>& q) {
            return var_keys[p.first] < var_keys[q.first];
        });
        ConjunctiveConditionEvaluator<V>* parent = NULL;
        ConjunctiveConditionEvaluator<V>* node = &result;
        for (unsigned j = 0; j < cond_sorted.size(); j++) {
            const auto& p = cond_sorted[j];
            if (node->m_var == p.first) {
                if (node->m_successors[p.second] == NULL) {
                    if (j + 1 < cond_sorted.size()) {
                        node->m_successors[p.second] =
                            new ConjunctiveConditionEvaluator<V>(
                                    cond_sorted[j+1].first,
                                    domains[cond_sorted[j+1].first]);
                    } else {
                        node->m_successors[p.second] =
                            new ConjunctiveConditionEvaluator<V>(-1, 0);
                    }
                }
                parent = node;
                node = node->m_successors[p.second];
            } else {
                if (node->m_var < 0
                        || var_keys[node->m_var] > var_keys[p.first]) {
                    assert(parent != NULL);
                    ConjunctiveConditionEvaluator<V>* node2 =
                            new ConjunctiveConditionEvaluator<V>(p.first, domains[p.first]);
                    if (j > 0 && cond_sorted[j - 1].first == parent->m_var) {
                        assert(parent->m_successors[cond_sorted[j - 1].second] == node);
                        parent->m_successors[cond_sorted[j - 1].second] = node2;
                    } else {
                        assert(parent->m_default == node);
                        parent->m_default = node2;
                    }
                    node2->m_default = node;
                    node = node2;
                } else {
                    if (node->m_default == NULL) {
                        node->m_default = 
                            new ConjunctiveConditionEvaluator<V>(p.first, domains[p.first]);
                    }
                    parent = node;
                    node = node->m_default;
                }
                j--;
            }
        }
        node->m_immediate.push_back(values[i]);
        cond_sorted.clear();
    }

    return result;
}

}
}

#endif
